/*
  Modul HTTP for ESP8266
  part of Arduino Mega Server project
*/

#ifdef HTTP_FEATURE

#include <ESP8266WebServer.h>

#define HTTP_PORT 8080
ESP8266WebServer HTTP(HTTP_PORT);

File uploadFile;

//блок переменных
#ifdef FILE_MANAGER
const uint16_t VERS_FILE_MANAGER = 1;
String filename = String();
long fileSize = 0;
uint32_t crcFile = crcInit;
unsigned long uptimeSec = 0;
#endif
//---------------

void httpInit() {
  initStart("HTTP");

  //блок инициализации запросов
#ifdef FILE_MANAGER
  Serial.println(" Version File Manager ESP: " + String(VERS_FILE_MANAGER));
  HTTP.on("/espfilelist", HTTP_GET, handleList);
  //HTTP.on("/test", HTTP_GET, handleTestCRC);
  HTTP.on("/esphash", HTTP_GET, handleGetHash);
  HTTP.on("/espupload", HTTP_POST, sendResponseOnPost, handleUpload);
  HTTP.on("/espman", HTTP_GET, handleAction);
  HTTP.on("/espdelete", HTTP_GET, handleDelete);
  HTTP.on("/espspiffsupdate", HTTP_GET, handleSpiffsUpdate);

  //  HTTP.on("/esprestart", HTTP_GET, []() {
  //    HTTP.send(200, "text/plain", "restart");
  //    ESP.restart();
  //  });
  //HTTP.on("/espupdate", HTTP_GET, handleDelete);
#endif
  //---------------------------

  HTTP.on("/list", HTTP_GET, handleFileList);
  HTTP.on("/edit", HTTP_GET, []() {
    if (!handleFileRead("/edit.htm")) HTTP.send(404, "text/plain", "FileNotFound");
  });

  HTTP.on("/edit", HTTP_PUT,    handleFileCreate);
  HTTP.on("/edit", HTTP_DELETE, handleFileDelete);

  //first callback is called after the request has ended with all parsed arguments
  //second callback handles file uploads at that location
  HTTP.on("/edit", HTTP_POST, []() {
    HTTP.send(200, "text/plain", "");
  }, handleFileUpload);

  // called when the url is not defined here, use it to load content from SPIFFS
  HTTP.onNotFound([]() {
    if (!handleFileRead(HTTP.uri()))
      HTTP.send(404, "text/plain", "FileNotFound");
  });

  updateInit();
  HTTP.begin();
  timeStamp();
  printValue("HTTP port", String(HTTP_PORT));
  initDone();
} // httpInit()

void httpWork() {
  HTTP.handleClient();
  delay(1);
}

String getContentType(String filename) {
  if (HTTP.hasArg("download"))         return "application/octet-stream";
  else if (filename.endsWith(".htm"))  return "text/html";
  else if (filename.endsWith(".html")) return "text/html";
  else if (filename.endsWith(".css"))  return "text/css";
  else if (filename.endsWith(".js"))   return "application/javascript";
  else if (filename.endsWith(".png"))  return "image/png";
  else if (filename.endsWith(".gif"))  return "image/gif";
  else if (filename.endsWith(".jpg"))  return "image/jpeg";
  else if (filename.endsWith(".ico"))  return "image/x-icon";
  else if (filename.endsWith(".xml"))  return "text/xml";
  else if (filename.endsWith(".pdf"))  return "application/x-pdf";
  else if (filename.endsWith(".zip"))  return "application/x-zip";
  else if (filename.endsWith(".gz"))   return "application/x-gzip";
  return "text/plain";
}

bool handleFileRead(String path) {
  if (path.endsWith("/")) path += "index.htm";
  String contentType = getContentType(path);
  String pathWithGz = path + ".gz";
  if (SPIFFS.exists(pathWithGz) || SPIFFS.exists(path)) {
    if (SPIFFS.exists(pathWithGz))
      path += ".gz";
    File file = SPIFFS.open(path, "r");
    size_t sent = HTTP.streamFile(file, contentType);
    file.close();
    return true;
  }
  return false;
}

void handleFileUpload() {
  if (HTTP.uri() != "/edit") return;
  HTTPUpload& upload = HTTP.upload();

  if (upload.status == UPLOAD_FILE_START) {
    String filename = upload.filename;
    if (!filename.startsWith("/")) filename = "/" + filename;
    uploadFile = SPIFFS.open(filename, "w");
    filename = String();
  }
  else if (upload.status == UPLOAD_FILE_WRITE) {
    //DBG_OUTPUT_PORT.print("handleFileUpload Data: ");
    //DBG_OUTPUT_PORT.println(upload.currentSize);
    if (uploadFile)
      uploadFile.write(upload.buf, upload.currentSize);
  }
  else if (upload.status == UPLOAD_FILE_END) {
    if (uploadFile)
      uploadFile.close();
  }
}

void handleFileDelete() {
  if (HTTP.args() == 0) return HTTP.send(500, "text/plain", "BAD ARGS");
  String path = HTTP.arg(0);
  Serial.println("handleFileDelete: " + path);
  if (path == "/")
    return HTTP.send(500, "text/plain", "BAD PATH");
  if (!SPIFFS.exists(path))
    return HTTP.send(404, "text/plain", "FileNotFound>" + path + "<");
  SPIFFS.remove(path);
  HTTP.send(200, "text/plain", "");
  path = String();
}

void handleFileCreate() {
  if (HTTP.args() == 0)
    return HTTP.send(500, "text/plain", "BAD ARGS");
  String path = HTTP.arg(0);
  if (path == "/")
    return HTTP.send(500, "text/plain", "BAD PATH");
  if (SPIFFS.exists(path))
    return HTTP.send(500, "text/plain", "FILE EXISTS");
  File file = SPIFFS.open(path, "w");
  if (file)
    file.close();
  else
    return HTTP.send(500, "text/plain", "CREATE FAILED");
  HTTP.send(200, "text/plain", "");
  path = String();
}

void handleFileList() {
  if (!HTTP.hasArg("dir")) {
    HTTP.send(500, "text/plain", "BAD ARGS");
    return;
  }
  String path = HTTP.arg("dir");
  Dir dir = SPIFFS.openDir(path);
  path = String();
  String output = "[";
  while (dir.next()) {
    File entry = dir.openFile("r");
    if (output != "[") output += ',';
    bool isDir = false;
    output += "{\"type\":\"";
    output += (isDir) ? "dir" : "file";
    output += "\",\"name\":\"";
    output += String(entry.name()).substring(1);
    output += "\"}";
    entry.close();
  }
  output += "]";
  HTTP.send(200, "text/json", output);
}// handleFileList()

void updateInit() {
  HTTP.on("/update", HTTP_POST, []() {
    HTTP.sendHeader("Connection", "close");
    HTTP.sendHeader("Access-Control-Allow-Origin", "*");
    HTTP.send(200, "text/plain", (Update.hasError()) ? "FAIL" : "OK");
    ESP.restart();
  }, []() {
    HTTPUpload& upload = HTTP.upload();
    if (upload.status == UPLOAD_FILE_START) {
      Serial.setDebugOutput(true);
      WiFiUDP::stopAll();
      Serial.printf("\n=== UPDATE PROCEDURE ==========\n%s\n", upload.filename.c_str());
      uint32_t maxSketchSpace = (ESP.getFreeSketchSpace() - 0x1000) & 0xFFFFF000;
      if (!Update.begin(maxSketchSpace)) { //start with max available size
        Update.printError(Serial);
      }
    }
    else if (upload.status == UPLOAD_FILE_WRITE) {
      if (Update.write(upload.buf, upload.currentSize) != upload.currentSize) {
        Update.printError(Serial);
      }
    }
    else if (upload.status == UPLOAD_FILE_END) {
      if (Update.end(true)) { //true to set the size to the current progress
        Serial.printf("Update Success: %u\n=== REBOOTING =================\n", upload.totalSize);
      } else {
        Update.printError(Serial);
      }
      Serial.setDebugOutput(false);
    }
    yield();
  });
} // updateInit()

///////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////функции AMS FileManager for ESP///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef FILE_MANAGER

void handleSpiffsUpdate() {
  if (!HTTP.hasArg("file")) return HTTP.send(500, "text/plain", "BAD ARGS");
  filename = HTTP.arg("file");
  if (filename == "/") return HTTP.send(500, "text/plain", "BAD PATH");
  if (!filename.startsWith("/")) filename = "/" + filename;
  if (!SPIFFS.exists(filename)) return HTTP.send(404, "text/plain", "FileNotFound>" + filename + "<");
  WiFiUDP::stopAll();
  Serial.println("\n=== UPDATE PROCEDURE ==========\n" + filename);
  uint32_t maxSketchSpace = (ESP.getFreeSketchSpace() - 0x1000) & 0xFFFFF000;
  if (!Update.begin(maxSketchSpace)) {
    Update.printError(Serial);
    return HTTP.send(500, "text/plain", "Error from the begining");
  }
  File file = SPIFFS.open(filename, "r");
  char buf[1024];
  int siz = file.size();
  while (siz > 0) {
    size_t len = std::min((int)(sizeof(buf) - 1), siz);
    file.read((uint8_t *)buf, len);
    buf[len] = 0;
    if (Update.write((uint8_t *)buf, len) != len) {
      Update.printError(Serial);
      //break;
    }
    siz -= sizeof(buf) - 1;
  }
  file.close();
  if (Update.end(true)) Serial.printf("Update Success: \n=== REBOOTING =================\n");
  else  Update.printError(Serial);
  yield();
  HTTP.sendHeader("Connection", "close");
//  HTTP.sendHeader("Access-Control-Allow-Origin", "*");
  HTTP.send(200, "text/plain", (Update.hasError()) ? "FAIL AND REBOOTING" : "OK AND REBOOTING");
  ESP.restart();
} // handleSpiffsUpdate()

void handleAction() {
  if (!HTTP.hasArg("action")) {
    HTTP.send(500, "text/plain", "BAD ARGS");
    return;
  }
  String action = HTTP.arg("action");
  if (action == "restart") {
    HTTP.send(200, "text/plain", "restarting");
    ESP.restart();
    return;
  }
  if (action == "format") {
    SPIFFS.format();
    Dir dir = SPIFFS.openDir("/");
    while (dir.next()) {
      HTTP.send(304, "text/plain", "not formated, file found");
      return;
    }
    HTTP.send(200, "text/plain", "formated");
    return;
  }
  if (action == "delete") {
    if (!HTTP.hasArg("file")) {
      HTTP.send(500, "text/plain", "BAD ARGS");
      return;
    }
    filename = HTTP.arg("file");
    if (!filename.startsWith("/")) filename = "/" + filename;
    if (filename == "/")
      HTTP.send(500, "text/plain", "BAD PATH");
    else if (!SPIFFS.exists(filename))
      HTTP.send(404, "text/plain", "FileNotFound>" + filename + "<");
    SPIFFS.remove(filename);
    if (!SPIFFS.exists(filename)) {
      HTTP.send(200, "text/plain", "deleted");
    } else {
      HTTP.send(304, "text/plain", "FileNotDelete>" + filename + "<");
    }
    filename = String();
    return;
  }
}
void sendResponseOnPost() {
  uint32_t fsize = HTTP.arg("size").toInt();
  uint32_t hash = HTTP.arg("hash").toInt();
  if (fileSize > 0) {
    getHash(filename);//запускаем генерацию crcFile
    String output = "{\"filename\":\"";
    output += filename;
    output += "\", \"size\":\"";
    output += String(fileSize);
    output += "\", \"hash\":\"";
    output += String(crcFile);
    output += "\"}";
    if (!HTTP.hasArg("size") || !HTTP.hasArg("hash")) {
      HTTP.send(500, "text/json", output);
    }
    if ((hash != crcFile) || (fsize != fileSize))
      HTTP.send(500, "text/json", output);
    else HTTP.send(200, "text/json", output);
  }
  else if (fileSize == fsize)
    HTTP.send(200, "text/plain", String(fileSize));
  else HTTP.send(304, "text/plain", String(fileSize));
  Serial.println("CRC(DEC): " + String(crcFile, DEC) + " size: " + String(fileSize) + " hash: " + String(crcFile) + " Time: " + String(millis() - uptimeSec));
  filename = String();
}

void handleUpload() {
  HTTPUpload& upload = HTTP.upload();
  if (upload.status == UPLOAD_FILE_START) {
    Serial.println("Start");
    uptimeSec = millis();
    filename = upload.filename;
    FSInfo fs_info;
    SPIFFS.info(fs_info);
    uint16_t fnLen = filename.length();
    if (fnLen > fs_info.maxPathLength - 1)
      filename = filename.substring(0, fs_info.maxPathLength - 5) + filename.substring(fnLen - 4);
    if (!filename.startsWith("/")) filename = "/" + filename;
    uploadFile = SPIFFS.open(filename, "w");
    fileSize = 0;
    crcFile = 0xFFFF;
    Serial.println("Upload file: " + filename);
  }
  else if (upload.status == UPLOAD_FILE_WRITE) {
    if (uploadFile) {
      uploadFile.write(upload.buf, upload.currentSize);
      //      crcFile = CRC_CCITT(upload.buf, upload.currentSize, crcFile);
    }
  }
  else if (upload.status == UPLOAD_FILE_END) {
    if (uploadFile) {
      if (uploadFile.size() == upload.totalSize) {
        fileSize = uploadFile.size();
      }
      else fileSize = -1;
      uploadFile.close();
      Serial.println("End");
    }
  }
}

void handleGetHash() {
  if (!HTTP.hasArg("file")) {
    HTTP.send(500, "text/plain", "BAD ARGS");
    return;
  }
  filename = HTTP.arg("file");
  if (!SPIFFS.exists(filename))
    return HTTP.send(404, "text/plain", "FileNotFound>" + filename + "<");
  getHash(filename);
  HTTP.send(200, "text/plain", String(crcFile));
  filename = String();
}

void getHash(String filename) {
  File file = SPIFFS.open(filename, "r");
  crcFile = 0xFFFF;
  char buf[1024];
  int siz = file.size();
  // Serial.println("Calc of file hash: "+filename);
  while (siz > 0) {
    size_t len = std::min((int)(sizeof(buf) - 1), siz);
    file.read((uint8_t *)buf, len);
    buf[len] = 0;
    //    Serial.print("buf: ");
    //    Serial.println(buf);
    crcFile = CRC_CCITT((uint8_t *)buf, len, crcFile);
    //      client.println(buf);
    siz -= sizeof(buf) - 1;
  }
  file.close();
}

void handleList() {
  long fileSize = 0;
  Dir dir = SPIFFS.openDir("/");
  String output = "{\"filelist\": [";
  bool first = true;
  while (dir.next()) {
    File entry = dir.openFile("r");
    fileSize += entry.size();
    if (!first) output += ',';
    else first = false;
    output += "{\"type\":\"file\", \"name\":\"";
    output += String(entry.name()).substring(1);
    output += "\",\"size\":";
    output += String(entry.size());
    output += "}";
    entry.close();
  }
  FSInfo fs_info;
  SPIFFS.info(fs_info);
  output += "], \"SPIFFSsize\": ";
  //output += String(ESP.getFlashChipRealSize() - (ESP.getSketchSize() + ESP.getFreeSketchSpace()));
  output += String(fs_info.totalBytes);
  output += ", \"fileSize\": ";
  //output += String(fileSize);
  output += String(fs_info.usedBytes);
  output += ", \"maxPathLength\": ";
  output += String(fs_info.maxPathLength);
  output += ", \"fmvers\": ";
  output += String(VERS_FILE_MANAGER);
  output += "}";

  HTTP.send(200, "text/json", output);
}

void handleDelete() {
  if (!HTTP.hasArg("file")) {
    HTTP.send(500, "text/plain", "BAD ARGS");
    return;
  }
  filename = HTTP.arg("file");
  Serial.println("handleFileDelete: " + filename);
  if (filename == "/")
    return HTTP.send(500, "text/plain", "BAD PATH");
  if (!SPIFFS.exists(filename))
    return HTTP.send(404, "text/plain", "FileNotFound>" + filename + "<");
  SPIFFS.remove(filename);
  HTTP.send(200, "text/plain", "deleted");
  filename = String();
}

/*void handleTestCRC() {
  if (!HTTP.hasArg("data")) {
    HTTP.send(500, "text/plain", "BAD ARGS");
    return;
  }
  String stringVar = HTTP.arg("data");
  char charVar[stringVar.length() + 1];
  stringVar.toCharArray(charVar, sizeof(charVar));

  Serial.print("word: " + stringVar);
  crcFile = 0xFFFF;
  crcFile = CRC_CCITT((uint8_t*)charVar, sizeof(charVar) - 1, crcFile);
  Serial.println(" key: " + String(crcFile) + " key(HEX): " + String(crcFile, HEX));
  Serial.println(crcFile, HEX);

  String output = "{\"data\":\"";
  output += stringVar;
  output += "\", \"hash\":\"";
  output += String(crcFile);
  output += "\"}";
  HTTP.send(200, "text/json", output);
  }*/
//--------------
#endif // FILE_MANAGER
#endif // HTTP_FEATURE
